from functools import lru_cache
import os
import sys
import time
import calendar
import warnings

import appdirs


LEAPFILE_NIST = '/usr/share/zoneinfo/leapseconds'
LEAPFILE_IETF = '/usr/share/zoneinfo/leap-seconds.list'
LEAPFILE_IETF_USER = os.path.join(
    appdirs.user_cache_dir('gpstime'), 'leap-seconds.list')
LEAPFILE_IETF_URL = 'https://www.ietf.org/timezones/data/leap-seconds.list'


def ntp2unix(ts):
    """Convert NTP timestamp to UTC UNIX timestamp

    1900-01-01T00:00:00Z -> 1970-01-01T00:00:00Z

    """
    return int(ts) - 2208988800


def load_NIST(path):
    data = []
    expires = 0
    with open(path, 'r') as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            if line[:8] == '#expires':
                expires = int(line.split()[1])
            elif line[0] == '#':
                continue
            else:
                year, mon, day, ts, correction = line.split()[1:6]
                st = time.strptime(
                    '{} {} {} {}'.format(year, mon, day, ts),
                    '%Y %b %d %H:%M:%S',
                )
                # FIXME: do something with correction
                data.append(calendar.timegm(st))
    return data, expires


def load_IETF(path):
    data = []
    expires = 0
    first = True
    with open(path, 'r') as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            elif line[:2] == '#@':
                expires = ntp2unix(line.split()[1])
            elif line[0] == '#':
                continue
            else:
                # ignore the first entry since that doesn't
                # actually correspond to a leap second
                if first:
                    first = False
                    continue
                leap, offset = line.split()[:2]
                # FIXME: do something with offset
                data.append(ntp2unix(leap))
    return data, expires


def fetch_ietf_leapfile(url=LEAPFILE_IETF_URL, path=LEAPFILE_IETF_USER):
    """Download IETF leap second data to path

    """
    import requests
    dd = os.path.dirname(path)
    if dd != '' and not os.path.exists(dd):
        os.makedirs(dd)
    r = requests.get(url)
    r.raise_for_status()
    tmp = path+'.tmp'
    with open(tmp, 'wb') as f:
        for c in r.iter_content():
            f.write(c)
    data, expires = load_IETF(tmp)
    os.rename(tmp, path)
    return data, expires


class LeapData:
    """Leap second data.

    """
    _GPS0 = 315964800

    def __init__(self):
        """Initialize leap second data

        """
        self._data = None
        self.expires = 0
        if os.path.exists(LEAPFILE_NIST):
            self._load(load_NIST, LEAPFILE_NIST)
        if not self.valid and os.path.exists(LEAPFILE_IETF):
            self._load(load_IETF, LEAPFILE_IETF)
        if not self.valid and os.path.exists(LEAPFILE_IETF_USER):
            self._load(load_IETF, LEAPFILE_IETF_USER)
        if not self.valid:
            if not self._data:
                print("Leap second data not available.", file=sys.stderr)
            elif self.expired:
                print("Leap second data is expired.", file=sys.stderr)
            print("Updating local user leap data cache from IETF...", file=sys.stderr)
            self._load(fetch_ietf_leapfile, LEAPFILE_IETF_URL)
            if not self._data:
                raise RuntimeError("Failed to load leap second data.")
            elif self.expired:
                warnings.warn("Leap second data is expired.", RuntimeWarning)

    def _load(self, func, path):
        try:
            self._data, self.expires = func(path)
        except:
            raise RuntimeError(f"Error loading leap file: {path}")

    @property
    def data(self):
        """Returns leap second data with times represented as UNIX.

        """
        if self.expired:
            warnings.warn("Leap second data is expired.", RuntimeWarning)
        return self._data

    @property
    def expired(self):
        """True if leap second data is expired

        """
        return self.expires <= time.time()

    @property
    def valid(self):
        """True if leap second data is available and not expired

        """
        return self._data and not self.expired

    def __iter__(self):
        for leap in self.data:
            yield leap

    @lru_cache(maxsize=None)
    def as_gps(self):
        """Returns leap second data with times represented as GPS.

        """
        leaps = [(leap - self._GPS0) for leap in self.data if leap >= self._GPS0]
        return [(leap + i) for i, leap in enumerate(leaps)]

    @lru_cache(maxsize=None)
    def as_unix(self, since_gps_epoch=False):
        """Returns leap second data with times represented as UNIX.

        If since_gps_epoch is set to True, only return leap second
        data since the GPS epoch (1980-01-06T00:00:00Z).

        """
        if since_gps_epoch:
            return [leap for leap in self.data if leap >= self._GPS0]
        else:
            return list(self.data)


LEAPDATA = LeapData()


if __name__ == '__main__':
    print("expires: {}".format(LEAPDATA.expires))
    for ls in LEAPDATA:
        print(ls)



import unittest
import sys
import datetime

from dateutil.tz import tzutc

import gpstime

##################################################

def cu_date_available():
    """return True if coreutils date parser available"""
    try:
        gpstime._cu_date_parse()
        return True
    except:
        return False


class TestGPStime(unittest.TestCase):

    def roundtrip_gps(self, gps):
        self.assertEqual(gpstime.unix2gps(gpstime.gps2unix(gps)), gps)

    def roundtrip_unix(self, unix):
        self.assertEqual(gpstime.gps2unix(gpstime.unix2gps(unix)), unix)

    ##########

    def test_conversion(self):
        self.assertEqual(gpstime.gps2unix(1133585676), 1449550459)
        self.assertEqual(gpstime.unix2gps(1449550459), 1133585676)
        self.assertEqual(1133585676, gpstime.unix2gps(gpstime.gps2unix(1133585676)))

    def test_conversion_past(self):
        self.assertEqual(gpstime.gps2unix(123456789), 439421586)
        self.assertEqual(gpstime.unix2gps(439421586), 123456789)

    def test_conversion_onleap(self):
        self.assertEqual(gpstime.gps2unix(425520006), 741484798)
        self.assertEqual(gpstime.gps2unix(425520007), 741484799)
        self.assertEqual(gpstime.gps2unix(425520008), 741484799)
        self.assertEqual(gpstime.gps2unix(425520009), 741484800)
        self.assertEqual(gpstime.unix2gps(741484798), 425520006)
        self.assertEqual(gpstime.unix2gps(741484799), 425520007)
        self.assertEqual(gpstime.unix2gps(741484800), 425520009)

    def test_roundtrip(self):
        self.roundtrip_gps(123456789)
        self.roundtrip_unix(439421586)

    @unittest.expectedFailure
    def test_roundtrip_onleap(self):
        self.roundtrip_gps(425520008)

    def test_gpstime_new(self):
        self.assertEqual(gpstime.gpstime(2015, 12, 8, 4, 54, 19, 0, tzutc()).gps(),
                         1133585676.0)

    def test_gpstime_fromdatetime(self):
        dt = datetime.datetime(2015, 12, 8, 4, 54, 19, 0, tzutc())
        gt = gpstime.gpstime.fromdatetime(dt)
        self.assertIsInstance(gt, gpstime.gpstime)
        self.assertEqual(gt.gps(), 1133585676.0)

    def test_gpstime_parse_now_roundtrip(self):
        gt0 = gpstime.gpstime.parse()
        gt1 = gpstime.gpstime.parse(gt0.gps())
        self.assertEqual(gt0, gt1)

    def test_gpstime_parse_utc(self):
        self.assertEqual(gpstime.gpstime.parse('Dec 08 2015 04:54:19.2 UTC').gps(),
                         1133585676.2)

    def test_gpstime_parse_utc2(self):
        self.assertEqual(gpstime.gpstime.parse('2014-07-03 17:16:14 UTC').gps(),
                         1088442990.0)

    def test_gpstime_parse_local(self):
        self.assertEqual(gpstime.gpstime.parse('Dec 08 2015 04:54:19.2 PST').gps(),
                         1133614476.2)

    def test_gpstime_parse_gps(self):
        self.assertEqual(gpstime.gpstime.parse(1133585676.2).gps(),
                         1133585676.2)

    def test_gpstime_parse_gps_iso(self):
        self.assertEqual(gpstime.gpstime.parse(1133585676.2).iso(),
                         '2015-12-08T04:54:19.200000Z')

    def test_gpstime_parse_iso(self):
        self.assertEqual(gpstime.gpstime.parse('2015-12-08T04:54:19.200000Z').iso(),
                         '2015-12-08T04:54:19.200000Z')

    def test_gpstime_parse_iso_alt(self):
        self.assertEqual(gpstime.gpstime.parse('2015-12-08_04:54:19.200000Z').iso(),
                         '2015-12-08T04:54:19.200000Z')

    @unittest.skipIf(not cu_date_available(), "coreutils date not available")
    def test_gpstime_parse_timestamp(self):
        self.assertEqual(gpstime.gpstime.parse('@1474821047').gps(),
                         1158856264)

    @unittest.skipIf(not cu_date_available(), "coreutils date not available")
    def test_gpstime_parse_now0(self):
        self.assertIsInstance(gpstime.gpstime.parse(), gpstime.gpstime)

    @unittest.skipIf(not cu_date_available(), "coreutils date not available")
    def test_gpstime_parse_now1(self):
        self.assertIsInstance(gpstime.gpstime.parse('now'), gpstime.gpstime)

    @unittest.skipIf(not cu_date_available(), "coreutils date not available")
    def test_gpstime_parse_relative0(self):
        self.assertIsInstance(gpstime.gpstime.parse('yesterday'), gpstime.gpstime)

    @unittest.skipIf(not cu_date_available(), "coreutils date not available")
    def test_gpstime_parse_relative1(self):
        self.assertIsInstance(gpstime.gpstime.parse('1 month ago'), gpstime.gpstime)

    def test_gpstime_parse_tconvert(self):
        self.assertEqual(gpstime.gpstime.tconvert('2015-12-08T04:54:19.200000Z'),
                         gpstime.gpstime.tconvert(1133585676.2))

    def test_gpstime_fromgps(self):
        self.assertEqual(gpstime.gpstime.fromgps(1133585676).iso(),
                         '2015-12-08T04:54:19.000000Z')

    def test_gpstime_fromgps_timestamp(self):
        self.assertEqual(gpstime.gpstime.fromgps(1133585676).timestamp(),
                         1449550459)

    def test_gpstime_tconvert_classmethod(self):
        self.assertEqual(gpstime.gpstime.tconvert(1133585676.2).gps(),
                         1133585676.2)

    def test_gpstime_tconvert_iso(self):
        self.assertEqual(gpstime.tconvert('2015-12-08T04:54:19.200000Z'),
                         1133585676.2)

    def test_gpstime_tconvert_gps(self):
        self.assertEqual(gpstime.tconvert(1133585676.2),
                         '2015-12-08 04:54:19.200000 UTC')

    @unittest.expectedFailure
    def test_gpstime_parse_leap(self):
        self.assertEqual(gpstime.gpstime.parse('Jun 30 1993 23:59:60 UTC').gps(),
                         425520008)

    @unittest.expectedFailure
    def test_gpstime_parse_gps_leap(self):
        self.assertEqual(gpstime.gpstime.parse(425520008).iso(),
                         '1993-06-30T23:59:60.000000Z')

    @unittest.skipIf(sys.version_info < (3, 8), "returns datetime in python < 3.8")
    def test_gpstime_add(self):
        dt = gpstime.gpstime.parse(1133585676.2)
        delta = datetime.timedelta(seconds=1)
        self.assertTrue(isinstance(dt + delta, gpstime.gpstime))
        dt += delta
        self.assertTrue(isinstance(dt, gpstime.gpstime))

    @unittest.skipIf(sys.version_info < (3, 8), "returns datetime in python < 3.8")
    def test_gpstime_subtract(self):
        dt = gpstime.gpstime.parse(1133585676.2)
        delta = datetime.timedelta(seconds=1)
        self.assertTrue(isinstance(dt - delta, gpstime.gpstime))
        dt -= delta
        self.assertTrue(isinstance(dt, gpstime.gpstime))

##################################################

if __name__ == '__main__':
    unittest.main(verbosity=5, failfast=False, buffer=True)



"""datetime with GPS time support

This module primarily provides functions for converting to/from UNIX
and GPS times, as well as a `gpstime` class that directly inherits
from the builtin `datetime` class, adding additional methods for GPS
time input and output.

Leap seconds are expected to be provided by the core libc Time Zone
Database tzdata.  If for some reason the tzdata leapsecond file is not
available, a local cache of the IETF leap second record will be
maintained:

  https://www.ietf.org/timezones/data/leap-seconds.list

KNOWN BUGS: This module does not currently handle conversions of time
strings describing the actual leap second themselves, which are
usually represented as the 60th second of the minute during which the
leap second occurs.

"""
import bisect
from datetime import datetime
import argparse
import subprocess

from dateutil.tz import tzutc, tzlocal

try:
    from .__version__ import version as __version__
except ModuleNotFoundError:
    try:
        import setuptools_scm
        __version__ = setuptools_scm.get_version(fallback_version='?.?.?')
    # FIXME: fallback_version is not available in the buster version
    # (3.2.0-1)
    except TypeError:
        __version__ = setuptools_scm.get_version()
    except LookupError:
        __version__ = '?.?.?'

from .leaps import LEAPDATA


##################################################

ISO_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'

# UNIX time for GPS 0 (1980-01-06T00:00:00Z)
GPS0 = 315964800

##################################################


class GPSTimeException(Exception):
    pass


def unix2gps(unix):
    """Convert UNIX timestamp to GPS time.

    """
    unix = float(unix)
    gps = unix - GPS0
    return gps + bisect.bisect(LEAPDATA.as_unix(since_gps_epoch=True), unix)


def gps2unix(gps):
    """Convert GPS time to UNIX timestamp.

    """
    gps = float(gps)
    unix = gps + GPS0
    return unix - bisect.bisect(LEAPDATA.as_gps(), gps)


##################################################


def _cu_date_parse(string='now'):
    """Parse date/time string to UNIX timestamp with GNU coreutils date

    """
    cmd = ['date', '+%s.%N', '-d', string]
    try:
        ts = subprocess.check_output(cmd, stderr=subprocess.PIPE).strip()
    except (FileNotFoundError, subprocess.CalledProcessError) as e:
        raise GPSTimeException("could not parse string '{}'".format(string)) from e
    return float(ts)


def _du_date_parse(string='now'):
    """Parse date/time string to UNIX timestamp with dateutils parse

    """
    from dateutil.parser import parse
    try:
        ts = parse(string).timestamp()
    except Exception as e:
        raise GPSTimeException("could not parse string '{}'".format(string)) from e
    return float(ts)


# test which date string parser to use.  default to coreutils date if
# available, since it has a better parser.
try:
    _cu_date_parse()
    _date_parse = _cu_date_parse
except GPSTimeException:
    _date_parse = _du_date_parse


##################################################


class gpstime(datetime):
    """GPS-aware datetime class

    An extension of the datetime class, with the addition of methods
    for converting to/from GPS times:

    >>> from gpstime import gpstime
    >>> gt = gpstime.fromgps(1088442990)
    >>> gt.gps()
    1088442990.0
    >>> gt.strftime('%Y-%m-%d %H:%M:%S %Z')
    '2014-07-03 17:16:14 UTC'
    >>> gpstime.now().gps()
    1133737481.204008

    In addition a natural language parsing `parse` classmethod returns
    a gpstime object for a arbitrary time string:

    >>> gpstime.parse('2014-07-03 17:16:14 UTC').gps()
    1088442990.0
    >>> gpstime.parse('2 days ago').gps()
    1158440653.553765

    """
    @classmethod
    def fromdatetime(cls, datetime):
        """Return gpstime object from datetime object"""
        tzinfo = datetime.tzinfo
        if tzinfo is None:
            tzinfo = tzlocal()
        return cls(
            datetime.year, datetime.month, datetime.day,
            datetime.hour, datetime.minute, datetime.second, datetime.microsecond,
            tzinfo,
        )

    @classmethod
    def fromgps(cls, gps):
        """Return gpstime object corresponding to GPS time."""
        gt = cls.utcfromtimestamp(gps2unix(gps))
        # HACK: in python3, utcfromtimestamp() seems to floor instead
        # of round the microseconds.  this causes round trips to fail.
        # manually fix microseconds here to the rounded value instead.
        ms = int(round(abs(gps - int(gps))*1000000))
        return gt.replace(microsecond=ms, tzinfo=tzutc())

    @classmethod
    def parse(cls, string='now'):
        """Parse an arbitrary time string into a gpstime object.

        If string not specified 'now' is assumed.  Strings that can be
        cast to float are assumed to be GPS times.  Prepend '@' to a
        float to specify a UNIX timestamp.

        This parse uses the natural lanuage parsing abilities of the
        GNU coreutils 'date' utility.  See "DATE STRING" in date(1)
        for information on possible date/time descriptions.

        """
        if string == 'now':
            return cls.now().replace(tzinfo=tzlocal())
        try:
            gps = float(string)
        except ValueError:
            gps = None
        except TypeError:
            raise TypeError("Time specification must be a string, not {!s}".format(type(string)))
        if gps is not None:
            return cls.fromgps(gps)
        try:
            ts = _date_parse(string)
        except GPSTimeException:
            # try again in case this was an ISO string using
            # underscore instead of T as the separator
            ts = _date_parse(string.replace('_', 'T'))
        return cls.fromtimestamp(ts).replace(tzinfo=tzlocal())

    tconvert = parse

    def gps(self):
        """Return GPS time as a float."""
        return unix2gps(self.timestamp())

    def iso(self):
        """Return time in standard UTC ISO format."""
        return self.astimezone(tzutc()).strftime(ISO_FORMAT)


def tconvert(string='now', form='%Y-%m-%d %H:%M:%S.%f %Z'):
    """Reimplementation of LIGO "tconvert" binary behavior

    Given a GPS time string, return the date/time string with the
    specified format.  Given a date/time string, return the GPS time.

    This just uses the gpstime.parse() method internally.

    """
    gt = gpstime.parse(string)
    try:
        float(string)
        return gt.strftime(form)
    except ValueError:
        return gt.gps()


def gpsnow():
    """Return current GPS time as a float.

    """
    return gpstime.utcnow().replace(tzinfo=tzutc()).gps()


parse = gpstime.parse


class GPSTimeParseAction(argparse.Action):
    """gpstime argparse argumention parser Action.

    Parses an arbitrary date/time string into a gpstime object.

    """
    def __call__(self, parser, namespace, values, option_string=False):
        # FIXME: support parsing argparse.REMAINDER values list into a
        # single string
        if isinstance(values, list):
            values = ' '.join(values)
        try:
            gps = parse(values)
        except GPSTimeException:
            parser.error("Could not parse date/time string '{}'".format(values))
        setattr(namespace, self.dest, gps)
